program hyp(input,output);

{This program generates a hypercube matrix 
which corresponds to hyp.radix.dim.dist of the report
R. H. Bisseling and W. F. McColl,
Scientific Computing on Bulk Synchronous Parallel Architectures,
Preprint 836, Department of Mathematics, Utrecht University,
dec. 1993.

Usage: hyp matfile}


const s1 = 200000; {maximum number of nonzeros}
      s2 = 3;      {maximum distance in graph}

var H : array[1..s2,0..s1,0..1] of integer;
    L,M : array[1..s2] of integer;
    x,k,d,p,q,r,error,v,rad,nplane,nall,n: integer;
    matrix : text;

function power(a,b:integer):integer;
{Pre:  b >= 0,
 Post: power = a**b}

begin
   if b=0 then power:=1
          else power:=a*power(a,b-1)
end; {power}



begin

    rewrite(matrix);
    open(matrix,'^1','new',error);
    
    writeln(output,'Input dimension >= 1');
    read(k);
    
    writeln(output,'Input radix > 1');
    read(rad);
    
    { mesh of size rad x rad x rad ... x rad, k times }
    
    n:= power(rad,k);
    {graph has n=rad**k vertices}
    
    d:=1;
    
    if rad=2 then
       begin
          writeln(output, 'Input distance >= 1');
          read(d)
       end;
    
    
    
    for q := 1 to d do
    begin
       H[q,0,0]:=0;
       H[q,0,1]:=0;
       L[q]:=1
    end;
    
    {H[q,cnt,0]=i and H[q,cnt,1]=j <=>
     there is an edge of digit distance <= q
     from node i to node j, registered at place cnt}
    
    for p := 1 to k do
       begin {a new dimension}
          nplane:= power(rad,p-1);
          nall:= power(rad,p);
          for q := 1 to d do
          begin
             {generate edges of duplicate planes}
             M[q]:=L[q];
             {L[q]=number of edges for dimension p-1 and distance <=q}
             for v := 1 to rad-1 do
                for r := 0 to (L[q]-1) do
                begin
                   H[q,v*L[q]+r,0]:=H[q,r,0] + v*nplane;
                   H[q,v*L[q]+r,1]:=H[q,r,1] + v*nplane;
                end;
             L[q]:=rad*L[q];
             {generate interplane edges}
             if q=1 then
                       {generate edges of dist 1 between planes}
                       for x:=0 to nall-1 do
                       begin
                            H[1,L[1],0]:=x;
                            H[1,L[1],1]:=(x+nplane) mod nall;
                            L[1]:=L[1]+1;
                            if rad>2 then 
                            begin
                               H[1,L[1],0]:=x;
                               H[1,L[1],1]:=(x+nall-nplane) mod nall;
                               L[1]:=L[1]+1
                            end;
                       end
                    else
                       for x := 0 to (M[q-1]-1) do
                       begin
                          H[q,L[q],0]:=H[q-1,x,0];
                          H[q,L[q],1]:=(H[q-1,x,1] + nplane) mod nall;
                          H[q,L[q]+1,0]:=H[q,L[q],1];
                          H[q,L[q]+1,1]:=H[q,L[q],0];
                          L[q]:=L[q]+2
                       end;  
             {L[q]=number of edges for dimension p and distance <=q}
          end;
       end; {of dimension}
    
    writeln(matrix,'S',n,n);
    
    for x := 0 to (L[d]-1) do writeln(matrix, H[d,x,0]:5,H[d,x,1]:5,1.1:4:1);
    writeln(matrix,-1);
    close(matrix);
    writeln(output,'n is',n);
    writeln(output,'m is',L[d])

end.
