program mkdhyp(input,output)

{This program generates a distribution
which corresponds to a hypercube matrix hyp.radix.dim.dist
of the report R. H. Bisseling and W. F. McColl,
Scientific Computing on Bulk Synchronous Parallel Architectures,
Preprint 836, Department of Mathematics, Utrecht University,
dec. 1993.

Usage: mkdhyp inpfile distrfile


{ This program creates a Cartesian distribution
  of a vector of length n over a q0 by q1 processor network,
  using an orthogonal partitioning of a hypercube graph of n vertices.

 Input: two filenames as arguments
        inpfile   = file containing
                    line 0       radix = number of vertices in one dimension,
                                 radix >= 1
                    line 1       dim = dimension of the hypercube, dim >= 1.
                    line 2       Pr[0] Pr[1] ... Pr[dim-1] 
                                 where Pr[k] = number of processor
                                               partitions in dimension k,
                                 Pr[k] is a divisor of radix, Pr[k] >= 1
                    line 3       q0 = number of processor rows
                    line 4       q1 = number of processor columns
                    (The total number of processors is
                                 q0*q1 = Pr[0]*Pr[1]* ... *Pr[dim-1])

        distrfile = new distribution file to be created 

Sample input file:
25
2
5 5
100
1
End sample input file


 Output: Distrfile contains the distribution in the following format
                    line 0              n q0 q1
                    line i+1            i D[i].s D[i].t for 0 <= i < n,
         Here, n = vector length = radix**dim,
         and the vector component i is allocated to 
         processor (D[i].s, D[i].t), with 0 <= D[i].s < q0
                                     and  0 <= D[i].t < q1. }

;const maxn   =  161000
;      maxq   =     100
;      maxdim =      20

;type pair   = record s,t : 0..maxq end
;type distr  = array[0..maxn]  of pair 
;type vec    = array[0..maxn]  of integer
;type dimvec = array[0..maxdim]  of integer

;var inpfile, distrfile: text

;procedure writedistr(D: distr; n,q0,q1: integer)
  ;var i: integer
  ;begin
      writeln(distrfile,n,q0,q1)
     ;for i:=0 to n-1 do writeln(distrfile,i,D[i].s,D[i].t)
   end {writedistr}

;procedure hyperdistr(var D0,D1: vec; Pr: dimvec; r,d,q0,n: integer)
  ;var k, i, rem, proc, p, pk, bk, ak: integer
  ;begin
    ;for i:=0 to n-1 do
       begin
          { i = (Sum l: 0 <= l < d : a[l] * (r**l) ),
            with 0 <= a[l] < r, all l. }
          p:=1
         ;proc:=0
         ;rem:= i
         ;for k:=0 to d-1 do
          begin
             { rem = (Sum l: k <= l < d : a[l] * (r**(l-k)) ) }
             ak:= rem mod r    
             { ak = a[k] }
            ;bk:= r div Pr[k]
             { bk = blocksize in dimension k } 
            ;pk:= ak div bk
             { 0 <= pk < Pr[k] }
            ;proc:= proc + pk*p
            ;p:= p*Pr[k]
            ;rem:= rem div r
          end
         ;D0[i]:= proc mod q0
         ;D1[i]:= proc div q0
       end
   end {hyperdistr}

;var D: distr 
;    D0,D1: vec
;    Pr: dimvec
;    i,radix,dim,k,prk,n,q0,q1,p: integer
;    error:integer

;begin {main}

    {******* Read matrix *******}
    ;open(inpfile,'^1','old',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening input file')
    ;reset(inpfile)

    ;open(distrfile,'^2','new',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening distribution file')

    ;readln(inpfile,radix)
    ;if radix < 1 then writeln(output, ' *** Error: radix < 1')
    ;readln(inpfile,dim)
    ;if dim < 1 then writeln(output, ' *** Error: dimension < 1')
    ;p:= 1
    ;n:=1
    ;for k:= 0 to dim-1 do
        begin
           read(inpfile,prk)
          ;Pr[k]:=prk
          ;if prk < 1 then writeln(output, ' *** Error: Pr-value < 1')
           else if (radix mod prk) <> 0 then
              writeln(output, ' *** Error: Pr-value is not a divisor of radix')
          ;p:= prk*p
          ;n:= radix*n
        end
    ;readln(inpfile)
    ;readln(inpfile,q0)
    ;if q0 < 1 then writeln(output, ' *** Error: q0 < 1')
    ;readln(inpfile,q1)
    ;if q1 < 1 then writeln(output, ' *** Error: q1 < 1')
    ;if p <> (q0*q1) then writeln(output, ' *** Error: p <> q0*q1 ')

    {******* Create distribution *******}
    ;hyperdistr(D0,D1,Pr,radix,dim,q0,n)

    {******* Write distribution *******}
    ;for i:= 0 to n-1 do
      begin
         D[i].s:= D0[i]
        ;D[i].t:= D1[i]
      end
    ;writedistr(D,n,q0,q1) 
    ;close(inpfile)
    ;close(distrfile)
end.

