program mvst (input,output)

{ This program gives the statistics of the sparse matrix-vector
  multiplication on a BSP architecture, for the BSP algorithm
  described in  R. H. Bisseling and W. F. McColl,
  Scientific Computing on Bulk Synchronous Parallel Architectures,
  Preprint 836, Department of Mathematics, Utrecht University,
  dec. 1993.

Usage: mvst matfile distrfile

 Input: two filenames as arguments
        matfile    = file containing the matrix in the following format
                    S n n              
                    (i j v)*
                    -1
                    where n = matrix size
                          i = row index of a matrix element
                          j = column index of a matrix element
                          v = value of a matrix element
         distrfile = file containing the distribution
                     in the following format
                     line 0              n q0 q1
                     line i+1            i D[i].s D[i].t for 0 <= i < n,
         Here vector component i is allocated to 
         processor (D[i].s, D[i].t), with 0 <= D[i].s < q0
                                    and  0 <= D[i].t < q1,
         and matrix element A[i,j] is allocated to 
         processor (D[i].s, D[j].t).

 Output: real numbers a, b, c, d, e on the standard output,
         representing the normalised BSP cost of the computation as
                a+c*g for the first superstep 
                b+d*g for the second superstep 
                e*l for the totall synchronisation cost.
         The total normalised BSP cost is (a+b) + (c+d)*g + e*l.

         The cost is normalised such that the cost of the perfectly 
         parallelised computation (i.e. without overhead)  is 1.}

;const maxnz =  1440000
;      maxn  =   160000
;      maxp  =      300

;type patpair  = record i,j : 0..maxn end
;type pair     = record s,t : 0..maxp end

;type sprspat = array[0..maxnz] of patpair
;type distr   = array[0..maxn]  of pair 
;type vec     = array[0..maxn]  of integer
;type count   = array[0..maxp] of integer

;var matfile, distrfile: text

;procedure readpat(var A: sprspat; var n,nz: integer)
  ;var x,m : integer
  ;    ch  : char
  ;    val : real
  ;begin
      ch:=chr(20)
     ;while ch<>'S' do read(matfile,ch)
     ;read(matfile,m,n)
     ; if m<>n then
         writeln(output, '*** Error: matrix should be square')
     ;nz:=0
     ;read(matfile,x)
     ;while x<>-1 do
      begin
         A[nz].i:=x
        ;read(matfile,A[nz].j,val,x)
        ;nz:=nz+1
      end
  end {readpat}

;procedure readdistr(var D: distr; var n,q0,q1: integer)
  ;var f, i, si, ti: integer
  ;begin
      readln(distrfile,n,q0,q1)
     ;for f:=0 to n-1 do
      begin
         readln(distrfile,i,si,ti)
        ;D[i].s:= si
        ;D[i].t:= ti
      end
   end {readdistr}

;procedure transposepat(var A: sprspat; nz: integer)
  ;var f, tmp: integer
  ;begin
      for f:= 0 to nz-1 do
         begin
            tmp:= A[f].i
           ;A[f].i:= A[f].j
           ;A[f].j:= tmp
         end
   end { transposepat }

;procedure transposedistr(var D: distr; n: integer)
  ;var f, tmp: integer
  ;begin
      for f:= 0 to n-1 do
         begin
            tmp:= D[f].s
           ;D[f].s:= D[f].t
           ;D[f].t:= tmp
         end
   end { transposedistr }

;procedure patsortcol(var A: sprspat; var C: vec;
                          n, nz: integer)
  {This subroutine sorts A by column, in situ}
  ;var f, j, src, dest: integer
  ;var tmp: patpair
  ;begin
      for j:=0 to n-1 do C[j]:= 0
     ;for f:=0 to nz-1 do
        begin
           j:=A[f].j
          ;C[j]:= C[j] + 1
        end
     {(A j: 0 <= j < n : C[j] = (N i: 0 <= i < n : A[i,j]<>0))}

     ;for j:=1 to n-1 do C[j]:= C[j-1] + C[j]
     ;C[n]:= nz 

     ;f:= 0
     ;while f < nz do
         begin
            j:=A[f].j
           ;if j=n then 
               {patpair A[f] is already marked as being
                in the correct position}
               f:=  f + 1
            else
               begin
                  src:= f
                 ;dest:= C[j]-1

                {swap A[src] and A[dest]}
                ;tmp:=A[dest]
                ;A[dest]:= A[src]
                ;A[src]:= tmp
            
                ;A[dest].j:= n
                ;C[j]:= C[j] - 1
              end
         end

     ;for j:=0 to n-1 do
        for f:=C[j] to C[j+1]-1 do
          A[f].j:= j
    
      {column j, 0 <= j < n, is placed in positions C[j]..C[j+1]-1} 
              
   end {patsortcol}

;procedure patsortrow(var A: sprspat; var C: vec;
                          n, nz: integer)
  {This subroutine sorts A by row, in situ}
  ;begin
      transposepat(A,nz)
     ;patsortcol(A,C,n,nz)
     ;transposepat(A,nz)

   end {patsortrow}

;procedure flops(var A: sprspat; var D: distr;
                     nz, q0, q1: integer;
                 var MaxFlops: integer)
  {Input: sparse pattern A, ordered by rows }
  ;var FlopCount, LastAccumulated: count
  ;    i, j, f, p, proc: integer
  {The two-dimensional processor identity (s,t) corresponds to the 
   one-dimensional identity proc by:   proc = t*q0 + s,
   where 0 <= s <  q0,    
         0 <= t <  q1,   
         0 <= proc < p = q0*q1 }  
  ;begin
      p:= q0*q1
     ;for proc:=0 to p-1 do 
          begin
             FlopCount[proc]:= 0
            ;LastAccumulated[proc]:= -1
          end
     ;for f:=0 to nz-1 do
        begin
           i:=A[f].i
          ;j:=A[f].j
          ;proc:= D[j].t*q0 + D[i].s
          ;if LastAccumulated[proc]=i then FlopCount[proc]:= FlopCount[proc] + 2
                                      else FlopCount[proc]:= FlopCount[proc] + 1
          ;LastAccumulated[proc]:=i
        end
     ;MaxFlops:=0
     ;for proc:=0 to p-1 do
        if FlopCount[proc]>MaxFlops then MaxFlops:= FlopCount[proc] 
   end {flops}

;procedure fanout(var A: sprspat; var D: distr; var C: vec;
                      n, q0, q1: integer;
                  var MaxHsend, MaxHrec, MaxFlops: integer)
  {Input: sparse pattern A, ordered by columns
          vector of pointers C, giving the starts of the columns of A}
  ;var i, j, p, proc, f, si, sj, tj, pij, pjj, jSent: integer
  ;var Hsend, Hrec, Flops, LastReceived: count
  ;var jLocallyNeeded: boolean
  ;begin
      p:= q0*q1
     ;for proc:=0 to p-1 do
           begin
              Hsend[proc]:= 0
             ;Hrec[proc]:= 0
             ;Flops[proc]:= 0
             ;LastReceived[proc]:= -1
           end
     ;for j:=0 to n-1 do 
        begin
           sj:= D[j].s
          ;tj:= D[j].t
          ;pjj:= tj*q0 + sj
          ;jLocallyNeeded:= false
          ;jSent:= 0
          ;for f:= C[j] to C[j+1]-1 do
             begin
                i:= A[f].i
               ;si:= D[i].s
               ;pij:= tj*q0 + si
               ;if (LastReceived[pij]<>j) and (si<>sj) then
                  begin
                      { component j is sent from processor (sj,tj)
                                            to   processor (si,tj) }
                      Hrec[pij]:= Hrec[pij]+1
                     ;LastReceived[pij]:= j
                     ;Hsend[pjj]:= Hsend[pjj] + 1
                     ;jSent:= jSent + 1
                  end
               ;if si=sj then jLocallyNeeded:= true
             end
           ;if jLocallyNeeded then Flops[pjj]:= Flops[pjj] + jSent
            else if jSent > 0 then Flops[pjj]:= Flops[pjj] + jSent - 1
        end            
     ;MaxHsend:=0
     ;MaxHrec:=0
     ;MaxFlops:= 0
     ;for proc:=0 to p-1 do
        begin
           if Hsend[proc]>MaxHsend then MaxHsend:= Hsend[proc] 
          ;if Hrec[proc] >MaxHrec then MaxHrec:= Hrec[proc] 
          ;if Flops[proc]>MaxFlops then MaxFlops:= Flops[proc] 
        end
   end {fanout}

;procedure fanin(var A: sprspat; var D: distr; var C: vec;
                     n, q0, q1: integer;
                 var MaxHsend, MaxHrec, MaxFlops: integer)
  {Input: sparse pattern A, ordered by rows
          vector of pointers C, giving the starts of the rows of A}
  ;var nz: integer
  ;begin
      nz:= C[n]
     ;transposepat(A,nz)
     ;transposedistr(D,n)
     ;fanout(A,D,C,n,q1,q0,MaxHrec,MaxHsend,MaxFlops)
     ;transposepat(A,nz)
     ;transposedistr(D,n)
   end {fanin}

;var A: sprspat
;    D: distr 
;    C: vec
;    j,n,nA,nz,q0,q1,p,error: integer
;    MaxFlops0, MaxFlops1, Nflops,
     MaxHsend0, MaxHsend1, MaxHrec0, MaxHrec1,
     MaxH0, MaxH1, dummy: integer
;    MinFlops, NormL: real

;begin {main}

    {******* Read matrix and distribution *******}
     open(matfile,'^1','old',error)
    ;if error<>0 then writeln(output,error,'*** Error: opening matrix file')
    ;open(distrfile,'^2','old',error)
    ;if error<>0 then writeln(output,error,'*** Error: opening distr file')
    ;reset(matfile)
    ;reset(distrfile)
    ;readpat(A,nA,nz)
    ;readdistr(D,n,q0,q1) 
    ;p:= q0*q1
    ;if n<>nA then
        writeln(output,'*** Error: distribution does not fit matrix size')

    ;patsortcol(A,C,n,nz)
    ;fanout(A,D,C,n,q0,q1,MaxHsend0,MaxHrec0,dummy)
    ;if MaxHsend0 > MaxHrec0 then MaxH0:= MaxHsend0 
                             else MaxH0:= MaxHrec0 
    ;patsortrow(A,C,n,nz)
    ;flops(A,D,nz,q0,q1,MaxFlops0)
    ;fanin(A,D,C,n,q0,q1,MaxHsend1,MaxHrec1,MaxFlops1)
    ;if MaxHsend1 > MaxHrec1 then MaxH1:= MaxHsend1 
                             else MaxH1:= MaxHrec1 

    { determine number of flops in sequential case }
    ;Nflops:= 0
    ;for j:=0 to n-1 do
       if C[j] < C[j+1] then Nflops:= Nflops + 2*(C[j+1]-C[j]) - 1
    ;MinFlops:= Nflops/p

    ;NormL:= 0.0
    ;if MaxH0>0 then NormL:= NormL +1
    ;if MaxH1>0 then NormL:= NormL +1
    ;if MaxFlops0>0 then NormL:= NormL +1
    ;if MaxFlops1>0 then NormL:= NormL +1
    ;if MinFlops>0.0 then NormL:= NormL/MinFlops
    ;writeln(output,MaxFlops0/MinFlops,MaxFlops1/MinFlops,
                    MaxH0/MinFlops,MaxH1/MinFlops, NormL)
end.
